/****************************************************************************/
/*                                                                          */
/* SERVER.C: All the data and code pertaining to the server are contained   */
/* in this module. It consists mostly of user interface control callbacks   */
/* where appropriate responses to user events are generated.                */
/*                                                                          */
/* The DDE services exported by server.c are listed at the top of server.h  */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/* Note:                                                                    */
/*      There are two problems with DDE under MS Windows that also          */
/*      appear when using the CVI DDE library.                              */
/*                                                                          */
/*      1)  The dataSize parameter passed to a DDE callback that is         */
/*          receiving data is not necessarily the same as the size of       */
/*          the data being transferred.  This value comes from Windows,     */
/*          and Windows seems to pass the size of its internal buffer       */
/*          instead of the actual data size.  This means that dataSize      */
/*          will as large or larger than the size of the data being         */
/*          transferred.  This also means that to transfer variable sized   */
/*          data, the size of the data must be encoded within the data      */
/*          itself (example: a NULL terminated string).                     */
/*                                                                          */
/*      2)  Windows implements DDE by posting messages to the message       */
/*          queues of the applications involved in the DDE conversation.    */
/*          Since these queues are small (limited by Windows) and since     */
/*          other events (mouse events, etc.) must use this queue as well,  */
/*          it is easy to overflow the queue by making continuous frequent  */
/*          DDE transfers.  When this happens, either DDE data will be lost */
/*          or corrupted, or mouse behavior will become erratic.  The       */
/*          work-around is to not make DDE transfers too frequently.  If    */
/*          a large amount of data must be transferred, send it in a few    */
/*          large chunks, rather than many small chunks.  This problem      */
/*          is easiest to observe when making multiple transfers less than  */
/*          a second apart on a slow or heavily loaded machine. On a fast   */
/*          PC, the problem is much less likely to occur.                   */
/*                                                                          */
/****************************************************************************/

/****************************************************************************/
/*                          Include Files                                   */
/****************************************************************************/

#include <ddesupp.h>    /* CF_TEXT, DDE_CONNECT, DDE_DISCONNECT, DDE_ADVISELOOP,
                           DDE_DATAREADY, DDE_REQUESTDATA, DDE_ADVISESTOP,
                           DDE_EXECUTE, ServerDDEWrite(), RegisterDDEServer(),
                           UnregisterDDEServer(), AdviseDDEDataReady() */
#include <userint.h>
#include <ansi_c.h>
#include "server.h"     /* TRUE, FALSE, NAME, TOPIC, ... */
#include "srvclt.h"     /* include file for the user-interface resource file */


/****************************************************************************/
/*                          Macro and Type Definitions                      */
/****************************************************************************/

typedef double  (*WaveFunctionPtr) (double radians);
#define PI      3.141592654


/****************************************************************************/
/*                          Function Prototypes                             */
/****************************************************************************/

static int CVICALLBACK     SrvCallback (unsigned int handle, char *topicName, char *itemName, int xType,
                    int dataFmt, int dataSize, void *dataPtr, void *callbackData);
static void     ExecuteCommandForClient(char *command);
static void     GenerateData(double cellArray[]);
static void     CopyDataToPanelCells(double cellArray[]);
static void     CopyDataFromPanelCells(double cellArray[]);
static double   Ramp(double radians);
static double   SawTooth(double radians);
static double   Random(double radians);


/****************************************************************************/
/*                          Global Variables                                */
/****************************************************************************/

int         panel;                  /* panel - server user interface panel handle */
int         hotLinkEnabled = FALSE; /* set to TRUE if hot link exists with a client on the cellValues[] data */
unsigned    convID;                 /* the server's conversation handle for the DDE connection */
double      cellValues[NUM_CELLS];  /* the array of data linked to by the client */

WaveFunctionPtr waveFunctions[] =   /* array of pointers to functions used for generating wave data */
    {sin, Ramp, SawTooth, Random};


/****************************************************************************/
/*                          Function Definitions                            */
/****************************************************************************/

    /*   Initializes this server program */
void ActivateServer (void)
{
    panel = LoadPanel (0, "srvclt.uir", SERVER);
    DisplayPanel (panel);           /* displays the server's user interface */

    RegisterDDEServer (NAME, SrvCallback, 0);   /* registers the program as a DDE server */
}

/****************************************************************************/

    /* Performs the server's termination tasks. */
void DeactivateServer (void)
{
    while (convID != -1)
        ProcessSystemEvents();  /* allows the server to receive the client's DDE_DISCONNECT callback before unregistering */
    UnregisterDDEServer (NAME); /* removes the program from the system's server database */
    DiscardPanel (panel);       /* discards the server's user interface */
}

/****************************************************************************/

    /*  Callback for the "Quit" button on the server panel */
int CVICALLBACK QuitCallback (int panel, int control, int event, void *callbackData,
    int eventData1, int eventData2)
{
    if (event == EVENT_COMMIT)
        QuitUserInterface (0);  /* terminates RunUserInterface() call inside main() */
    return FALSE;
}

/****************************************************************************/

    /*  DDE server callback. This function receives and handles all messages
        sent by clients. The type of message is determined by the 'xType'
        parameter.
    */
static int CVICALLBACK SrvCallback (unsigned int handle, char *topicName, char *itemName, int xType,
                int dataFmt, int dataSize, void *dataPtr, void *callbackData)
{
    double  value;
    int     cellIndex;      /* 0 refers to CELL 1, ... , 19 refers to CELL 20 */

    if (xType != DDE_CONNECT && handle != convID ||
        xType != DDE_DISCONNECT && strcmp (topicName, TOPIC))
        return FALSE;           /* checks message validity. Only acknowledges
                                   messages pertaining to one topic (TOPIC)
                                   (except for the disconnect request (DDE_DISCONNECT)
                                   where no topic is sent), and one conversation
                                   (convID) (except for the connection request
                                   (DDE_CONNECT) where the conversation handle isn't
                                   yet defined) */

    switch (xType)      /* determines DDE message type */
        {
        case DDE_CONNECT:
            convID = handle;    /* assigns new conversation handle to convID */
            break;

        case DDE_DISCONNECT:
            convID = -1;        /* eliminates old conversation handle */
            break;

        case DDE_ADVISELOOP:
            if (strcmp(itemName, ALL_CELLS) != 0)
                return FALSE;   /* this server only supports a hot link for the array of cells as a whole */
            hotLinkEnabled = TRUE;        /* establish new link */
            break;

        case DDE_DATAREADY:
            if ((strncmp(itemName, CELL, strlen(CELL)) != 0)
                || dataFmt != CF_TEXT)
                return FALSE;   /* checks item and data format validity */
            cellIndex = atoi(itemName + strlen(CELL)) - 1;  /* Given a string "CELLx" (where x is a number from 1 to 20), skip the "CELL" characters and get the value of x (and then subtract one). */
            CopyDataFromPanelCells(cellValues);
            cellValues[cellIndex] = *(double *)dataPtr;
            CopyDataToPanelCells(cellValues);
            break;              /* stores data received from client in specified cell control */

        case DDE_REQUESTDATA:
            if ((strncmp (itemName, CELL, strlen(CELL)) != 0)|| dataFmt != CF_TEXT)
                return FALSE;   /* checks item and data format validity */
            cellIndex = atoi(itemName + strlen(CELL)) - 1;  /* Given a string "CELLx" (where x is a number from 1 to 20), skip the "CELL" characters and get the value of x (and then subtract one). */
            CopyDataFromPanelCells(cellValues);
            ServerDDEWrite (convID, itemName, CF_TEXT, &cellValues[cellIndex], sizeof(cellValues[cellIndex]),
                DDE_TIMEOUT);          /* obtains data from specified cell control and sends it to client */
            break;

        case DDE_ADVISESTOP:
            if (strcmp (itemName, ALL_CELLS) !=0)
                return FALSE;   /* hot links are only supported on one item (knob) */
            hotLinkEnabled = FALSE;       /* remove link */
            break;

        case DDE_EXECUTE:
            ExecuteCommandForClient(itemName);
            break;

        default:
            return FALSE;       /* unrecognized client message */
        }

    return TRUE;                /* TRUE --> server processed message successfully */
}

/*****************************************************************************/

    /*  Executes a DDE command sent from a client. All commands are executed
        by activating the command button which performs the command and
        faking a space bar keystroke to operate the button.  This method
        makes it easier for the user to see what is happening, which is
        appropriate since this is a demonstration program.
    */
static void ExecuteCommandForClient(char *command)
{
    int control;

    if (strcmp(command, CMD_GENERATED_DATA) == 0)
        control = SERVER_GENERATENEWDATA;
    else
    if (strcmp(command, CMD_ADD_DATA) == 0)
        control = SERVER_ADDTOCURRENTDATA;
    else
    if (strcmp(command, CMD_SHIFT_LEFT) == 0)
        control = SERVER_SHIFTDATALEFT;
    else
    if (strcmp(command, CMD_SHIFT_RIGHT) == 0)
        control = SERVER_SHIFTDATARIGHT;
    else
    if (strcmp(command, CMD_QUIT) == 0)
        control = SERVER_QUIT;
    else
        return;     /* Ignore unrecognized command */

    SetActivePanel(panel);
    SetActiveCtrl(panel, control);      /* activate the command button for the specified command */
    FakeKeystroke(' ');                 /* pretend the user hit the space bar on the button, to make it operate */
}

/*****************************************************************************/

    /*  Fills in an array of doubles with a waveform controlled by the
        amplitude knob and the wave shape slider.
    */
static void GenerateData(double cellArray[])
{
    int     index;
    int     functionIndex;
    double  amplitude;

    GetCtrlVal(panel, SERVER_WAVESHAPE, &functionIndex);    /*  get the index of which wave function to use */
    GetCtrlVal(panel, SERVER_AMPLITUDE, &amplitude);        /*  get amplitude value to scale the waveform by */

    for (index = 0; index < NUM_CELLS; index++) /* generate one complete cycle of the waveform */
        cellArray[index]  = amplitude * waveFunctions[functionIndex](index * 2 * PI / NUM_CELLS);
}

/*****************************************************************************/

    /*  This function is the function used to change the data in the numerics
        on the server panel programmatically.  It sets the value of each
        numeric according to the values passed in cellArray, and it notifies
        the client of the new data.
    */
static void CopyDataToPanelCells(double cellArray[])
{
    int index;

        /*  This function can get the control id of each numeric by adding to the
            control id of the first numeric because the numeric controls are
            arrange in ascending tab order on the panel.
        */
    for (index = 0; index < NUM_CELLS; index++)
        SetCtrlAttribute(panel, SERVER_CELL1 + index, ATTR_CTRL_VAL, cellArray[index]); /* used ATTR_CTRL_VAL instead of calling SetCtrlVal() since it is faster setting the value of several overlapped controls at the same time */

    if (hotLinkEnabled)
        AdviseDDEDataReady (convID, ALL_CELLS, CF_TEXT, cellArray,  sizeof (cellValues), DDE_TIMEOUT);  /* notifies client */
}

/*****************************************************************************/

    /*  This function copies the data from the numerics (cells) on the
        server panel in the cellArray passed in.
    */
static void CopyDataFromPanelCells(double cellArray[])
{
    int index;

        /*  This function can get the control id of each numeric by adding to the
            control id of the first numeric because the numeric controls are
            arrange in ascending tab order on the panel.
        */
    for (index = 0; index < NUM_CELLS; index++)
        GetCtrlVal(panel, SERVER_CELL1 + index, &cellArray[index]);
}

/*****************************************************************************/

    /*  Callback for the "Generate New Data" button. It creates a new waveform
        and copies it into the numerics on the server panel.
    */
int CVICALLBACK GenerateDataCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    if (event == EVENT_COMMIT)
        {
        GenerateData(cellValues);
        CopyDataToPanelCells(cellValues);
        }

    return FALSE;
}

/*****************************************************************************/

    /*  Callback for the "Add To Current Data" button.  It creates a new
        waveform and then adds it to the waveform values already in the
        numerics on the server panel.
    */
int CVICALLBACK AddToDataCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    int     index;
    double  waveToAdd[NUM_CELLS];

    if (event == EVENT_COMMIT)
        {
        GenerateData(waveToAdd);            /*  get new waveform */
        CopyDataFromPanelCells(cellValues); /*  copy current waveform from the numerics on the panel */

            /*  add both waveforms */
        for (index = 0; index < NUM_CELLS; index++)
            cellValues[index] += waveToAdd[index];
        CopyDataToPanelCells(cellValues);   /*  copy 'added' waveform back to the numerics on the server panel */
        }

    return FALSE;
}

/*****************************************************************************/

    /*  Callback for the "Shift Data Left" button.  Shifts the waveform in the
        numerics on the server panel down one element, wrapping the first
        element back to the end of the wave.
    */
int CVICALLBACK ShiftDataLeftCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    int     index;
    double  firstValue;

    if (event == EVENT_COMMIT)
        {
        CopyDataFromPanelCells(cellValues);     /* get the current waveform on the panel */
        firstValue = cellValues[0];             /* save the first value */

            /* move every value down one element in the waveform */
        for (index = 0; index < NUM_CELLS - 1; index++)
            cellValues[index] = cellValues[index + 1];
        cellValues[NUM_CELLS - 1] = firstValue; /* place the first value at the end */
        CopyDataToPanelCells(cellValues);       /* copy the waveform back to the numerics on the server panel */
        }

    return FALSE;
}

/*****************************************************************************/

    /*  Callback for the "Shift Data Right" button.  Shifts the waveform in the
        numerics on the server panel up one element, wrapping the last
        element back to the start of the wave.
    */
int CVICALLBACK ShiftDataRightCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    int     index;
    double  lastValue;

    if (event == EVENT_COMMIT)
        {
        CopyDataFromPanelCells(cellValues);     /* get the current waveform on the panel */
        lastValue = cellValues[NUM_CELLS - 1];  /* save the last value */

            /* move every value up one element in the waveform */
        for (index = NUM_CELLS-1; index > 0; index--)
            cellValues[index] = cellValues[index - 1];
        cellValues[0] = lastValue;              /* place the last value at the start */
        CopyDataToPanelCells(cellValues);       /* copy the waveform back to the numerics on the server panel */
        }

    return FALSE;
}

/*****************************************************************************/

    /*  This is the callback for EVERY numeric on the server panel.  If
        a new value is entered in any of the numerics (by the user
        editing the value). The client is notified about the new data.
    */
int CVICALLBACK CellCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    if (event == EVENT_VAL_CHANGED)
        if (hotLinkEnabled)
            {
            CopyDataFromPanelCells(cellValues);
            AdviseDDEDataReady (convID, ALL_CELLS, CF_TEXT, cellValues,  sizeof (cellValues), DDE_TIMEOUT); /* notifies client */
            }

    return FALSE;
}

/*****************************************************************************/

    /*  Generates a point in a ramp waveform */
static double Ramp(double radians)
{
    return fmod(radians, 2 * PI) / PI - 1;
}

/*****************************************************************************/

    /*  Generates a point in a sawtooth waveform */
static double SawTooth(double radians)
{
    double  val = 2 * fmod(radians, 2 * PI) / PI - 1;

    if (val > 1)
        val = 2 - val;  /* negative slope on second half of cycle */

    return val;
}

/*****************************************************************************/

    /*  Generates a point in a random waveform */
static double Random(double radians)
{
    return -2 *(double)rand() / RAND_MAX + 1;
}





