/****************************************************************************/
/* CVISVR.C: All the data and code pertaining to the server are contained   */
/* in this module. It consists mostly of user interface control callbacks   */
/* where appropriate responses to user events are generated.                    */
/****************************************************************************/


/****************************************************************************/
/*                          Include Files                                   */
/****************************************************************************/

#include <ddesupp.h>    /* CF_TEXT, DDE_CONNECT, DDE_DISCONNECT, DDE_ADVISELOOP,
                           DDE_DATAREADY, DDE_REQUESTDATA, DDE_ADVISESTOP,
                           DDE_EXECUTE, ServerDDEWrite(), RegisterDDEServer(),
                           UnregisterDDEServer(), AdviseDDEDataReady() */
#include <userint.h>    /* ATTR_MIN_VALUE, ATTR_MAX_VALUE, EVENT_VAL_CHANGED,
                           EVENT_COMMIT, LoadPanel(), DiscardPanel(),
                           SetCtrlVal(), GetCtrlVal(), GetCtrlAttribute(),
                           QuitUserInterface(), ProcessSystemEvents(),
                           DisplayPanel() */
#include <ansi_c.h>     /* strcmp(), RAND_MAX, rand() */
#include "cvisvr.h"     /* TRUE, FALSE, NAME, TOPIC, KNOB_ITEM, COLOR_ITEM,
                           COMMAND, WAIT */
#include "cvidde.h"     /* SERVER, SERVER_AUTO, SERVER_CONTROL, SERVER_QUIT,
                           SERVER_INOUT */

/****************************************************************************/
/*                          Macro Definitions                               */
/****************************************************************************/

#define MAX_STEP    3   /* increment range of the data control in auto mode */

/****************************************************************************/
/*                          Global Variables                                */
/****************************************************************************/

static int          panel, autoMode, flow = FALSE;
    /* panel - user interface panel handle */
    /* autoMode - boolean: when TRUE, server is in auto mode, where data changes continuously */
    /* flow - boolean: when TRUE, a hot link exists with a client, on the data */
static unsigned int convID;
    /* convID - the server's conversation handle for the DDE connection */

/****************************************************************************/
/*                          Function Prototypes                             */
/****************************************************************************/

int CVICALLBACK DataCallback (int panel, int control, int event, void *callbackData,
        int eventData1, int eventData2);
int CVICALLBACK AutoCallback (int panel, int control, int event, void *callbackData,
        int eventData1, int eventData2);
int CVICALLBACK QuitCallback (int panel, int control, int event, void *callbackData,
        int eventData1, int eventData2);
int CVICALLBACK SrvCallback (unsigned int handle, char *topicName, char *itemName, int xType,
        int dataFmt, int dataSize, void *dataPtr, void *callbackData);

/****************************************************************************/
/*                          Function Definitions                            */
/****************************************************************************/

/****************************************************************************/
/* Performs the server's initialization tasks.                              */
/****************************************************************************/
void ActivateServer (void)
{
    panel = LoadPanel (0, "cvidde.uir", SERVER);
    DisplayPanel (panel);
    GetCtrlVal (panel, SERVER_AUTO, &autoMode); /* obtains initial auto mode status */

    RegisterDDEServer (NAME, SrvCallback, 0);   /* registers program as a DDE server */

    return;
}


/****************************************************************************/
/* Performs the server's termination tasks.                                 */
/****************************************************************************/
void DeactivateServer (void)
{
    UnregisterDDEServer (NAME); /* removes program from the system's server database */
    DiscardPanel (panel);

    return;
}


/****************************************************************************/
/* Performs the series of tasks in which the server needs to engage, on     */
/* each idle event, when placed in auto mode: namely, generate new data and */
/* if a client has set up a hot link (advisory loop) for that data, notify  */
/* the client.                                                              */
/****************************************************************************/
void ProcessServerIdleEvent (void)
{
    int min, max, newVal, oldVal;
    char buffer[100]={0};

    if (autoMode) {                         /* obtains current data range */
        GetCtrlAttribute (panel, SERVER_CONTROL, ATTR_MIN_VALUE, &min);
        GetCtrlAttribute (panel, SERVER_CONTROL, ATTR_MAX_VALUE, &max);
        GetCtrlVal (panel, SERVER_CONTROL, &oldVal);

                                            /* generates new data */
        newVal = oldVal + (2 * MAX_STEP + 1) * rand() / RAND_MAX - MAX_STEP;

                                            /* insures that data is in range */
        if (newVal != oldVal && newVal >= min && newVal <= max) {
            SetCtrlVal (panel, SERVER_CONTROL, newVal); /* updates data control */
            if (flow) {                     /* TRUE --> link exists */
                sprintf(buffer, "%d\0", newVal);
                AdviseDDEDataReady (convID, KNOB_ITEM, CF_TEXT, buffer,
                    strlen(buffer)+1, WAIT);    /* notifies client */
                ProcessSystemEvents();      /* allows DDE Management Library
                                               the opportunity to process
                                               request, before a new one is
                                               submitted on the next idle event */
                }
            }
        }

    return;
}


/****************************************************************************/
/* Callback for the data control (knob). Only responds to VAL_CHANGE events.*/
/****************************************************************************/
int CVICALLBACK DataCallback (int panel, int control, int event, void *callbackData,
    int eventData1, int eventData2)
{
    int data;
    char buffer[10]={0};

    if (control != SERVER_CONTROL)                  /* checks callback validity */
        return FALSE;

    if (event == EVENT_VAL_CHANGED && flow) {       /* checks existence of link */
        GetCtrlVal (panel, SERVER_CONTROL, &data);  /* gets new data value */
        sprintf(buffer, "%d\0", data);
        AdviseDDEDataReady (convID, KNOB_ITEM, CF_TEXT, buffer, strlen (buffer)+1,
            WAIT);                                  /* notifies the client */
        }

    return FALSE;
}


/****************************************************************************/
/* Callback for the termination control (command button). Only responds to  */
/* COMMIT events.                                                           */
/****************************************************************************/
int CVICALLBACK QuitCallback (int panel, int control, int event, void *callbackData,
    int eventData1, int eventData2)
{
    if (control != SERVER_QUIT) /* checks callback validity */
        return FALSE;

    if (event == EVENT_COMMIT)
        QuitUserInterface (0);  /* terminates RunUserInterface() call inside main program */
    return FALSE;
}


/****************************************************************************/
/* Callback for the auto mode control (radio button). Only responds to      */
/* COMMIT events.                                                           */
/****************************************************************************/
int CVICALLBACK AutoCallback (int panel, int control, int event, void *callbackData,
    int eventData1, int eventData2)
{
    if (control != SERVER_AUTO)                     /* checks callback validity */
        return FALSE;

    if (event == EVENT_COMMIT)
        GetCtrlVal (panel, SERVER_AUTO, &autoMode); /* obtains new auto mode status */

    return FALSE;
}



/****************************************************************************/
/* DDE server callback. This function receives and handles all messages     */
/* sent by clients. The type of message is determined by the 'xType'        */
/* parameter.                                                               */
/****************************************************************************/
int CVICALLBACK SrvCallback (unsigned int handle, char *topicName, char *itemName, int xType,
    int dataFmt, int dataSize, void *dataPtr, void *callbackData)
{
    int  data;
    char buffer[100]={0};

    if (xType != DDE_CONNECT && handle != convID || strcmp (topicName, TOPIC))
        return FALSE;           /* checks message validity. Only acknowledges
                                   messages pertaining to one topic (TOPIC) and
                                   one conversation (convID) (except for the
                                   connection request (DDE_CONNECT) where the
                                   conversation handle isn't yet defined) */

    switch (xType) {            /* determines message type */
        case DDE_CONNECT:
            convID = handle;    /* assigns new conversation handle to convID */
            break;
        case DDE_DISCONNECT:
            convID = -1;        /* eliminates old conversation handle */
            break;
        case DDE_ADVISELOOP:
            if (strcmp (itemName, KNOB_ITEM))
                return FALSE;   /* hot links are only supported on one item (knob) */
            flow = TRUE;        /* establish new link */
            break;
        case DDE_DATAREADY:
            MessagePopup ("Message from the VB Client", (char *)dataPtr);
            break;              /* display message sent by the Visual Basic Client */
        case DDE_REQUESTDATA:
            if (strcmp (itemName, KNOB_ITEM) || dataFmt != CF_TEXT)
                return FALSE;   /* checks item and data format validity */
            GetCtrlVal (panel, SERVER_CONTROL, &data);  /* gets new data value */
            sprintf(buffer, "%d\0", data);
            ServerDDEWrite (convID, KNOB_ITEM, CF_TEXT, &buffer, strlen (buffer)+1,
                WAIT);          /* obtains data from color control and sends it to client */
            break;
        case DDE_ADVISESTOP:
            if (strcmp (itemName, KNOB_ITEM))
                return FALSE;   /* hot links are only supported on one item (knob) */
            flow = FALSE;       /* remove link */
            break;
        case DDE_EXECUTE:
            if (strcmp (itemName, COMMAND))
                return FALSE;   /* checks command validity. Only one type is
                                   supported (COMMAND) */
            GetCtrlVal (panel, SERVER_AUTO, &autoMode);
            autoMode = !autoMode;
            SetCtrlVal (panel, SERVER_AUTO, autoMode);
            break;              /* performs command: toggles auto mode status */
        default:
            return FALSE;       /* unrecognized client message */
        }

    return TRUE;                /* TRUE --> server processed message successfully */
}
